// This file is part of Blend2D project <https://blend2d.com>
//
// See blend2d.h or LICENSE.md for license and copyright information
// SPDX-License-Identifier: Zlib

// ----------------------------------------------------------------------------
// Blend2D - 2D Vector Graphics Powered by a JIT Compiler.
//
//  * Official Blend2D Home Page: https://blend2d.com
//  * Official Github Repository: https://github.com/blend2d/blend2d
//
// This header is provided for users that want to extend Blend2D library. It
// provides some utilities that make such extending easier.
// ----------------------------------------------------------------------------

#ifndef BLEND2D_IMPL_H_INCLUDED
#define BLEND2D_IMPL_H_INCLUDED

#ifndef __cplusplus
  #error "blend2d-impl.h header can only be included in C++ mode"
#endif

#include "blend2d.h"
#include "blend2d/api-impl.h"

#endif // BLEND2D_IMPL_H_INCLUDED
